We will create a server that when visited appends a timestamp in a `visited.log` file on the host computer (otherwise if it's the container, that container is **ephemeral** and the visited.log can disappear when you restart the container or the container crashes)

Let's run a NodeJS express server at port 3002 on the host computer so that visiting `http://localhost:3002/` will show a json message "message": "Logged timestamp"

And when visiting`http://localhost:3002/visited`, will show all the timestamps of hitting the `/` page. Alternately, you could open `visited.log` file to see the timestamps.

This assumes you have knowledge of 1.06 on how to run a NodeJS Express server off of Docker image.

---

.dockerignore:
```
.DS_Store
node_modules
package-lock.json
__pycache__
*.pyc
.env
```

Dockerfile:
```
# Use the specific NodeJS base image (bundled in Debian)
FROM node:18

# Set the working directory in the container.
# All below commands will work from working directory.
WORKDIR /app

# Copy the package.json file to the container
COPY package.json .

# Install the NodeJS packages
RUN npm install

# Copy your application files (if any)
COPY . .

# Specify the command to run your application (optional, depends on your use case)
CMD ["node", "app.js"]
```

package.json:
```
{
  "dependencies": {
    "express": "4.21.2"
  }
}
```

app.js:
```
const express = require('express');
const fs = require('fs');
const path = require('path');
const app = express();

// Ensure logs directory exists
const LOGS_DIR = './logs';
if (!fs.existsSync(LOGS_DIR)) {
    fs.mkdirSync(LOGS_DIR, { recursive: true });
}
const LOG_FILE = path.join(LOGS_DIR, 'visited.log');

app.get('/', (req, res) => {
    const timestamp = new Date().toISOString();
    fs.appendFileSync(LOG_FILE, `${timestamp}\n`);
    res.json({ message: "Logged timestamp" });
});

app.get('/visited', (req, res) => {
    try {
        const logs = fs.readFileSync(LOG_FILE, 'utf8');
        res.send(`<pre>${logs}</pre>`);
    } catch (error) {
        if (error.code === 'ENOENT') {
            res.send('No visits logged yet');
        } else {
            res.status(500).send('Error reading logs');
        }
    }
});

const port = 3000;
app.listen(port, () => {
  console.log(`Server running on port ${port}`);
});
```
^ The key is that the path built for LOG_FILE is "./logs/visited.log"

logs/visited.log:
```
```
^ Create a logs/ folder. Optionally you could create an empty visited.log file.


---

**Test directly**


At this point, test the `app.js` directly:
```
node app.js
```

Note this will be port 3000 because we haven't mapped to port 3002 via `docker run` command that mapped the container's port to the host's port.

Now visit in the web browser `http://localhost:3000/` directly to log a timestamp. Refresh a few times.

Now visit in the web browser `http://localhost:3000/visited` or open the file `logs/visited.log` to see logged timestamps.

Close app.js instance. We will move to image/container's in the next section

---

**Image/Container with Mounted Path**

When you build the image, then a virtual container off the image, you wouldn't want visited.log to be created inside the virtual container, because it's ephemeral (data gets wiped if the container crashes or restarts).

Let's build the image (make sure to run this command where the Dockerfile and logs/ folder, and the other files are at):
```
docker build -t express-app .
```

Then we run it on port 3002 or your choice of port, and mount the paths as follows:
```
docker run -p 3002:3000 -v $(pwd)/logs:/app/logs -d express-app
```

^FUNDAMENTA: Just like the ports follow the order of HOST:CONTAINER, so does the mounting path. 

^FUNDAMENTAL: By having a `logs/` folder at where the Dockerfile is, the COPY command in the Dockerfile makes sure there's a `logs/` folder in the container too because it copies all files and folders over when running `docker build..`. However, the container `logs/` folder becomes a symbolic link to your host's `logs/` folder. Therefore, the containerized app reads and writes symbolically to `/app/logs` (Because `WORKDIR app/`) when in actuality, it reads and writes to your host's `logs/` folder.  The `$(pwd)` is a shortcut way of building the absolute path to the host's logs/ folder where you ran the `docker build..` command.

^Again, -d places the process into the background in daemon mode (aka detached mode in Docker terminology). So the terminal does not hang until you close out the server. When you're done, cleanup includes running `docker ps` to get the container id to kill with `docker kill CONTAINER_ID`

>[!note] FYI on Terminology
> - The paths can be described as: the `logs` directory from the host (`$(pwd)/logs`) is mounted to the `/app/logs` directory inside the container.
> - **"Map"** is often used for **port binding** (`-p 3002:3000`), but **"Mount"** is the correct term because Docker is attaching the host directory (`$(pwd)/logs`) to the container’s file system (`/app/logs`) in a way that allows data persistence.


Now visit in the web browser `http://localhost:3002/` directly to log a timestamp. Refresh a few times.

Now visit in the web browser `http://localhost:3002/visited` or open the file `logs/visited.log` to see logged timestamps.

If all is working as expected, this concludes the challenge.